// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

TEXT errors(SB),$0
	CSRRC	(X10), CYCLE, X5		// ERROR "integer register or immediate expected for 1st operand"
	CSRRC	X0, TU, X5			// ERROR "unknown CSR"
	CSRRC	X0, CYCLE			// ERROR "missing CSR name"
	CSRRC	X0, CYCLE, (X10)		// ERROR "needs an integer register output"
	CSRRC	$-1, TIME, X15			// ERROR "immediate out of range 0 to 31"
	CSRRCI	$32, TIME, X15			// ERROR "immediate out of range 0 to 31"
	CSRRCI	$1, TIME, (X15)			// ERROR "needs an integer register output"
	CSRRS	(X10), CYCLE, X5		// ERROR "integer register or immediate expected for 1st operand"
	CSRRS	X0, CYCLE, (X10)		// ERROR "needs an integer register output"
	CSRRS	X0, TU, X5			// ERROR "unknown CSR"
	CSRRS	X0, CYCLE			// ERROR "missing CSR name"
	CSRRS	$-1, TIME, X15			// ERROR "immediate out of range 0 to 31"
	CSRRSI	$32, TIME, X15			// ERROR "immediate out of range 0 to 31"
	CSRRSI	$1, TIME, (X15)			// ERROR "needs an integer register output"
	CSRRW	(X10), CYCLE, X5		// ERROR "integer register or immediate expected for 1st operand"
	CSRRW	X0, TU, X5			// ERROR "unknown CSR"
	CSRRW	X0, CYCLE			// ERROR "missing CSR name"
	CSRRW	X0, CYCLE, (X5)			// ERROR "needs an integer register output"
	CSRRW	$-1, TIME, X15			// ERROR "immediate out of range 0 to 31"
	CSRRWI	$32, TIME, X15			// ERROR "immediate out of range 0 to 31"
	CSRRWI	$1, TIME, (X15)			// ERROR "needs an integer register output"
	MOV	$errors(SB), (X5)		// ERROR "address load must target register"
	MOV	$8(SP), (X5)			// ERROR "address load must target register"
	MOVB	$8(SP), X5			// ERROR "unsupported address load"
	MOVH	$8(SP), X5			// ERROR "unsupported address load"
	MOVW	$8(SP), X5			// ERROR "unsupported address load"
	MOVF	$8(SP), X5			// ERROR "unsupported address load"
	MOV	$1234, 0(SP)			// ERROR "constant load must target register"
	MOV	$1234, 8(SP)			// ERROR "constant load must target register"
	MOV	$0, 0(SP)			// ERROR "constant load must target register"
	MOV	$0, 8(SP)			// ERROR "constant load must target register"
	MOV	$1234, 0(SP)			// ERROR "constant load must target register"
	MOV	$1234, 8(SP)			// ERROR "constant load must target register"
	MOVB	$1, X5				// ERROR "unsupported constant load"
	MOVH	$1, X5				// ERROR "unsupported constant load"
	MOVW	$1, X5				// ERROR "unsupported constant load"
	MOVF	$1, X5				// ERROR "unsupported constant load"
	MOVBU	X5, (X6)			// ERROR "unsupported unsigned store"
	MOVHU	X5, (X6)			// ERROR "unsupported unsigned store"
	MOVWU	X5, (X6)			// ERROR "unsupported unsigned store"
	MOVF	F0, F1, F2			// ERROR "illegal MOV instruction"
	MOVD	F0, F1, F2			// ERROR "illegal MOV instruction"
	MOV	X10, X11, X12			// ERROR "illegal MOV instruction"
	MOVW	X10, X11, X12			// ERROR "illegal MOV instruction"
	RORI	$64, X5, X6			// ERROR "immediate out of range 0 to 63"
	SLLI	$64, X5, X6			// ERROR "immediate out of range 0 to 63"
	SRLI	$64, X5, X6			// ERROR "immediate out of range 0 to 63"
	SRAI	$64, X5, X6			// ERROR "immediate out of range 0 to 63"
	BEQ	X5, X6, $1			// ERROR "instruction with branch-like opcode lacks destination"
	BEQ	X5, X6, 31(X10)			// ERROR "instruction with branch-like opcode lacks destination"
	RORI	$-1, X5, X6			// ERROR "immediate out of range 0 to 63"
	SLLI	$-1, X5, X6			// ERROR "immediate out of range 0 to 63"
	SRLI	$-1, X5, X6			// ERROR "immediate out of range 0 to 63"
	SRAI	$-1, X5, X6			// ERROR "immediate out of range 0 to 63"
	RORIW	$32, X5, X6			// ERROR "immediate out of range 0 to 31"
	SLLIW	$32, X5, X6			// ERROR "immediate out of range 0 to 31"
	SRLIW	$32, X5, X6			// ERROR "immediate out of range 0 to 31"
	SRAIW	$32, X5, X6			// ERROR "immediate out of range 0 to 31"
	RORIW	$-1, X5, X6			// ERROR "immediate out of range 0 to 31"
	SLLIW	$-1, X5, X6			// ERROR "immediate out of range 0 to 31"
	SRLIW	$-1, X5, X6			// ERROR "immediate out of range 0 to 31"
	SRAIW	$-1, X5, X6			// ERROR "immediate out of range 0 to 31"
	SD	X5, 4294967296(X6)		// ERROR "constant 4294967296 too large"
	FNES	F1, (X5)			// ERROR "needs an integer register output"

	//
	// "V" Standard Extension for Vector Operations, Version 1.0
	//
	VSETIVLI	X10, E32, M2, TA, MA, X12	// ERROR "expected immediate value"
	VLE8V		(X10), V1, V3			// ERROR "invalid vector mask register"
	VLE8FFV		(X10), V1, V3			// ERROR "invalid vector mask register"
	VSE8V		V3, V1, (X10)			// ERROR "invalid vector mask register"
	VLSE8V		(X10), X10, V1, V3		// ERROR "invalid vector mask register"
	VSSE8V		V3, X11, V1, (X10)		// ERROR "invalid vector mask register"
	VLUXEI8V	(X10), V2, V1, V3		// ERROR "invalid vector mask register"
	VSUXEI8V	V3, V2, V1, (X10)		// ERROR "invalid vector mask register"
	VLOXEI8V	(X10), V2, V1, V3		// ERROR "invalid vector mask register"
	VSOXEI8V	V3, V2, V1, (X10)		// ERROR "invalid vector mask register"
	VLSEG2E8V	(X10), V1, V3			// ERROR "invalid vector mask register"
	VLSEG2E8FFV	(X10), V1, V3			// ERROR "invalid vector mask register"
	VSSEG2E8V	V3, V1, (X10)			// ERROR "invalid vector mask register"
	VLSSEG2E8V	(X10), X10, V1, V3		// ERROR "invalid vector mask register"
	VSSSEG2E8V	V3, X11, V1, (X10)		// ERROR "invalid vector mask register"
	VLUXSEG2EI8V	(X10), V2, V1, V3		// ERROR "invalid vector mask register"
	VSUXSEG2EI8V	V3, V2, V1, (X10)		// ERROR "invalid vector mask register"
	VLOXSEG2EI8V	(X10), V2, V1, V3		// ERROR "invalid vector mask register"
	VSOXSEG2EI8V	V3, V2, V1, (X10)		// ERROR "invalid vector mask register"
	VL1RV		(X10), V0, V3			// ERROR "too many operands for instruction"
	VS1RV		V3, V0, (X11)			// ERROR "too many operands for instruction"
	VADDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VADDVX		X10, V2, V1, V3			// ERROR "invalid vector mask register"
	VADDVI		$15, V4, V1, V2			// ERROR "invalid vector mask register"
	VSUBVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSUBVX		X10, V2, V1, V3			// ERROR "invalid vector mask register"
	VRSUBVX		X10, V2, V1, V3			// ERROR "invalid vector mask register"
	VRSUBVI		$15, V4, V1, V2			// ERROR "invalid vector mask register"
	VNEGV		V2, V3, V4			// ERROR "invalid vector mask register"
	VWADDUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDUWV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDUWX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBUWV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBUWX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDWV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWADDWX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBWV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWSUBWX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWCVTXXV	V2, V1, V3			// ERROR "invalid vector mask register"
	VWCVTUXXV	V2, V1, V3			// ERROR "invalid vector mask register"
	VZEXTVF2	V2, V3, V4			// ERROR "invalid vector mask register"
	VSEXTVF2	V2, V3, V4			// ERROR "invalid vector mask register"
	VZEXTVF4	V2, V3, V4			// ERROR "invalid vector mask register"
	VSEXTVF4	V2, V3, V4			// ERROR "invalid vector mask register"
	VZEXTVF8	V2, V3, V4			// ERROR "invalid vector mask register"
	VSEXTVF8	V2, V3, V4			// ERROR "invalid vector mask register"
	VADCVVM		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VADCVVM		V1, V2, V3			// ERROR "invalid vector mask register"
	VADCVVM		V1, V2, V0, V0			// ERROR "invalid destination register V0"
	VADCVXM		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VADCVXM		X10, V2, V3			// ERROR "invalid vector mask register"
	VADCVXM		X10, V2, V0, V0			// ERROR "invalid destination register V0"
	VADCVIM		$15, V2, V1, V3			// ERROR "invalid vector mask register"
	VADCVIM		$15, V2, V3			// ERROR "invalid vector mask register"
	VADCVIM		$15, V2, V0, V0			// ERROR "invalid destination register V0"
	VMADCVVM	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMADCVVM	V1, V2, V3			// ERROR "invalid vector mask register"
	VMADCVXM	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMADCVXM	X10, V2, V3			// ERROR "invalid vector mask register"
	VMADCVIM	$15, V2, V1, V3			// ERROR "invalid vector mask register"
	VMADCVIM	$15, V2, V3			// ERROR "invalid vector mask register"
	VSBCVVM		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSBCVVM		V1, V2, V3			// ERROR "invalid vector mask register"
	VSBCVVM		V1, V2, V0, V0			// ERROR "invalid destination register V0"
	VSBCVXM		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSBCVXM		X10, V2, V3			// ERROR "invalid vector mask register"
	VSBCVXM		X10, V2, V0, V0			// ERROR "invalid destination register V0"
	VMSBCVVM	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSBCVVM	V1, V2, V3			// ERROR "invalid vector mask register"
	VMSBCVXM	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSBCVXM	X10, V2, V3			// ERROR "invalid vector mask register"
	VANDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VANDVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VANDVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VORVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VORVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VORVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VXORVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VXORVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VXORVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VNOTV		V1, V2, V3			// ERROR "invalid vector mask register"
	VSLLVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLLVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLLVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VSRLVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSRLVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSRLVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VSRAVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSRAVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSRAVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VNSRLWV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VNSRLWX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VNSRLWI		$31, V2, V4, V3			// ERROR "invalid vector mask register"
	VNSRAWV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VNSRAWX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VNSRAWI		$31, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCVTXXW	V2, V4, V3			// ERROR "invalid vector mask register"
	VMSEQVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSEQVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSEQVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSNEVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSNEVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSNEVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLTUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLTUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLTVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLTVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLEUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLEUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLEUVI	$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLEVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLEVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLEVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGTUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGTUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGTUVI	$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGTVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGTVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGTVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGEVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGEUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLTVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSLTUVI	$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGEVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMSGEUVI	$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMINUVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMINUVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMINVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMINVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMAXUVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMAXUVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMAXVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMAXVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULHVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULHVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULHUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULHUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULHSUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMULHSUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VDIVUVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VDIVUVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VDIVVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VDIVVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VREMUVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREMUVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VREMVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREMVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMULVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMULVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMULUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMULUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMULSUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMULSUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMACCVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMACCVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VNMSACVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VNMSACVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMADDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMADDVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VNMSUBVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VNMSUBVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCSUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCSUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VWMACCUSVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMERGEVVM	V1, V2, V3			// ERROR "invalid vector mask register"
	VMERGEVVM	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMERGEVXM	X10, V2, V3			// ERROR "invalid vector mask register"
	VMERGEVXM	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMERGEVIM	$15, V2, V3			// ERROR "invalid vector mask register"
	VMERGEVIM	$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VMVVV		V1, V2, V3			// ERROR "too many operands for instruction"
	VMVVX		X10, V2, V3			// ERROR "too many operands for instruction"
	VMVVI		$15, V2, V3			// ERROR "too many operands for instruction"
	VSADDUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSADDUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSADDUVI	$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VSADDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSADDVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSADDVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSUBUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSUBUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSUBVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSUBVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VAADDUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VAADDUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VAADDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VAADDVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VASUBUVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VASUBUVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VASUBVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VASUBVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSMULVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSMULVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSRLVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSRLVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSRLVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSRAVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSRAVX		X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSSRAVI		$15, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCLIPUWV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCLIPUWX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCLIPUWI	$16, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCLIPWV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCLIPWX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VNCLIPWI	$16, V2, V4, V3			// ERROR "invalid vector mask register"
	VFADDVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFADDVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSUBVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSUBVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFRSUBVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWADDVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWADDVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWSUBVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWSUBVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWADDWV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWADDWF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWSUBWV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWSUBWF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMULVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMULVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFDIVVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFDIVVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFRDIVVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWMULVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWMULVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMACCVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFMACCVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFNMACCVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFNMACCVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFMSACVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFMSACVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFNMSACVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFNMSACVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFMADDVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFMADDVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFNMADDVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFNMADDVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFMSUBVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFMSUBVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFNMSUBVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFNMSUBVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFWMACCVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFWMACCVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFWNMACCVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFWNMACCVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFWMSACVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFWMSACVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFWNMSACVV	V2, V1, V4, V3			// ERROR "invalid vector mask register"
	VFWNMSACVF	V2, F10, V4, V3			// ERROR "invalid vector mask register"
	VFSQRTV		V2, V4, V3			// ERROR "invalid vector mask register"
	VFRSQRT7V	V2, V4, V3			// ERROR "invalid vector mask register"
	VFREC7V		V2, V4, V3			// ERROR "invalid vector mask register"
	VFMINVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMINVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMAXVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMAXVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSGNJVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSGNJVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSGNJNVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSGNJNVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSGNJXVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSGNJXVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFNEGV		V2, V4, V3			// ERROR "invalid vector mask register"
	VFABSV		V2, V4, V3			// ERROR "invalid vector mask register"
	VMFEQVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFEQVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFNEVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFNEVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFLTVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFLTVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFLEVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFLEVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFGTVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFGEVF		F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFGTVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VMFGEVV		V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFMERGEVFM	X10, V2, V3			// ERROR "invalid vector mask register"
	VFMERGEVFM	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFCVTXUFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFCVTXFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFCVTRTZXUFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFCVTRTZXFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFCVTFXUV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFCVTFXV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTXUFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTXFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTRTZXUFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTRTZXFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTFXUV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTFXV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFWCVTFFV	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTXUFW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTXFW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTRTZXUFW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTRTZXFW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTFXUW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTFXW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTFFW	V2, V4, V3			// ERROR "invalid vector mask register"
	VFNCVTRODFFW	V2, V4, V3			// ERROR "invalid vector mask register"
	VREDSUMVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDMAXUVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDMAXVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDMINUVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDMINVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDANDVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDORVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VREDXORVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWREDSUMUVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VWREDSUMVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFREDOSUMVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFREDUSUMVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFREDMAXVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFREDMINVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VFWREDOSUMVS	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VCPOPM		V2, V4, X10			// ERROR "invalid vector mask register"
	VFIRSTM		V2, V4, X10			// ERROR "invalid vector mask register"
	VMSBFM		V2, V4, V3			// ERROR "invalid vector mask register"
	VMSIFM		V2, V4, V3			// ERROR "invalid vector mask register"
	VMSOFM		V2, V4, V3			// ERROR "invalid vector mask register"
	VIOTAM		V2, V4, V3			// ERROR "invalid vector mask register"
	VSLIDEUPVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLIDEUPVI	$16, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLIDEDOWNVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLIDEDOWNVI	$16, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLIDE1UPVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSLIDE1UPVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VSLIDE1DOWNVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VFSLIDE1DOWNVF	F10, V2, V4, V3			// ERROR "invalid vector mask register"
	VRGATHERVV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VRGATHEREI16VV	V1, V2, V4, V3			// ERROR "invalid vector mask register"
	VRGATHERVX	X10, V2, V4, V3			// ERROR "invalid vector mask register"
	VRGATHERVI	$16, V2, V4, V3			// ERROR "invalid vector mask register"

	RET
