// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build fips140v1.0

package fipstest

import "testing"

func fips140v126Conditionals() {}

func testFIPS140v126(t *testing.T, plaintext []byte) {}
