// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Part 2 of the sources for issue 57046 test case.
// See part 1 for build instructions.

const char *mom() {
  return "mom";
}
