// -lang=go1.20

// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package p

func _(s []int) {
	clear /* ERROR "clear requires go1.21 or later" */ (s)
}
