// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package p

func _()         { f(g) }
func f[P any](P) {}
func g[Q int](Q) {}
